var ts = require('typescript');
var extend = require('xtend');
var has = require('has');
var Promise = require('native-or-bluebird');
var os_1 = require('os');
var path_1 = require('path');
var fs_1 = require('../utils/fs');
var path_2 = require('../utils/path');
var references_1 = require('../utils/references');
function compile(tree, options) {
    var files = {};
    return Promise.all([
        compileDependencyTree(tree, extend(options, { browser: false, files: files })),
        compileDependencyTree(tree, extend(options, { browser: true, files: files }))
    ])
        .then(function (_a) {
        var main = _a[0], browser = _a[1];
        return ({ main: main, browser: browser });
    });
}
Object.defineProperty(exports, "__esModule", { value: true });
exports.default = compile;
function getStringifyOptions(tree, options) {
    var overrides = {};
    var isTypings = typeof tree.typings === 'string';
    var main = isTypings ? tree.typings : tree.main;
    var browser = isTypings ? tree.browserTypings : tree.browser;
    if (options.browser && browser) {
        if (typeof browser === 'string') {
            var mainDefinition = path_2.resolveFrom(tree.src, path_2.normalizeToDefinition(main));
            var browserDefinition = path_2.resolveFrom(tree.src, path_2.normalizeToDefinition(browser));
            overrides[mainDefinition] = browserDefinition;
        }
        else {
            var browserOverrides = browser;
            for (var _i = 0, _a = Object.keys(browserOverrides); _i < _a.length; _i++) {
                var key = _a[_i];
                if (!path_2.isModuleName(key)) {
                    var from = path_2.resolveFrom(tree.src, path_2.normalizeToDefinition(key));
                    var to = path_2.resolveFrom(tree.src, path_2.normalizeToDefinition(browserOverrides[key]));
                    overrides[from] = to;
                }
                else {
                    overrides[key] = browserOverrides[key];
                }
            }
        }
    }
    var imported = {};
    var dependencies = {};
    var entry = path_2.resolveFrom(tree.src, path_2.normalizeToDefinition(main));
    return extend(options, {
        tree: tree,
        entry: entry,
        isTypings: isTypings,
        overrides: overrides,
        imported: imported,
        dependencies: dependencies
    });
}
function compileDependencyTree(tree, options) {
    return compileDependencyPath(null, getStringifyOptions(tree, options));
}
function compileDependencyPath(path, options) {
    var tree = options.tree, entry = options.entry;
    return stringifyDependencyPath(path_2.resolveFrom(tree.src, path == null ? entry : path), options);
}
function cachedReadFileFrom(path, options) {
    if (!has(options.files, path)) {
        options.files[path] = fs_1.readFileFrom(path);
    }
    return options.files[path];
}
function cachedStringifyOptions(name, compileOptions, options) {
    var tree = getDependency(name, options);
    if (!has(options.dependencies, name)) {
        if (tree) {
            options.dependencies[name] = getStringifyOptions(tree, compileOptions);
        }
        else {
            options.dependencies[name] = null;
        }
    }
    return options.dependencies[name];
}
function getDependency(name, options) {
    var tree = options.tree, overrides = options.overrides;
    if (has(options.overrides, name)) {
        return tree.dependencies[overrides[name]];
    }
    if (has(tree.dependencies, name)) {
        return tree.dependencies[name];
    }
    if (options.strict) {
        throw new TypeError("Unable to resolve \"" + name + "\" from \"" + options.name + "\"");
    }
}
function stringifyDependencyPath(path, options) {
    var definitionPath = path_2.normalizeToDefinition(path);
    if (has(options.overrides, definitionPath)) {
        definitionPath = options.overrides[definitionPath];
    }
    return cachedReadFileFrom(definitionPath, options)
        .then(function (contents) {
        var info = ts.preProcessFile(contents);
        var tree = options.tree, ambient = options.ambient, cwd = options.cwd, browser = options.browser, name = options.name, files = options.files;
        var ambientModules = info.ambientExternalModules || [];
        if (info.isLibFile) {
            return;
        }
        if (ambientModules.length && !ambient) {
            return Promise.reject(new TypeError("Attempted to compile " + name + " as non-ambient when it contains external module declarations"));
        }
        var importedFiles = info.importedFiles.map(function (x) { return path_2.isModuleName(x.fileName) ? x.fileName : path_2.resolveFrom(path, x.fileName); });
        var referencedFiles = info.referencedFiles.map(function (x) { return path_2.resolveFrom(path, x.fileName); });
        if (ambient) {
            Object.keys(tree.dependencies).forEach(function (x) { return importedFiles.push(x); });
        }
        var imports = importedFiles.map(function (path) {
            if (has(options.imported, path)) {
                return;
            }
            if (ambientModules.indexOf(path) > -1) {
                return;
            }
            options.imported[path] = true;
            if (path_2.isModuleName(path)) {
                var parts = path.split('/');
                var dependencyName = parts.shift();
                var dependencyPath = parts.length === 0 ? null : parts.join('/');
                var moduleName = ambient ? dependencyName : name + "!" + dependencyName;
                var compileOptions = { cwd: cwd, browser: browser, files: files, name: moduleName, ambient: false };
                var stringifyOptions = cachedStringifyOptions(dependencyName, compileOptions, options);
                if (!stringifyOptions) {
                    return;
                }
                return compileDependencyPath(dependencyPath, stringifyOptions);
            }
            return stringifyDependencyPath(path, options);
        });
        return Promise.all(imports)
            .then(function (files) {
            files.push(stringifyFile(path, contents.replace(references_1.REFERENCE_REGEXP, ''), options));
            return files.filter(function (x) { return x != null; }).join(os_1.EOL);
        });
    });
}
function stringifyFile(path, contents, options) {
    var sourceFile = ts.createSourceFile(path, contents, 2, true);
    var tree = options.tree, name = options.name;
    var isES6Export = true;
    var wasDeclared = false;
    if (options.ambient) {
        if (sourceFile.externalModuleIndicator) {
            throw new TypeError("Unable to compile \"" + path + "\" - looks like an external module");
        }
        return contents.trim();
    }
    function importPath(name) {
        if (path_2.isModuleName(name)) {
            return options.name + "!" + name;
        }
        var modulePath = path_2.relativeTo(tree.src, path_2.resolveFrom(path, name));
        return path_2.normalizeSlashes(path_1.join(options.name, modulePath));
    }
    function replacer(node) {
        if (node.kind === 225) {
            isES6Export = !node.isExportEquals;
        }
        if (node.kind === 9 &&
            (node.parent.kind === 226 ||
                node.parent.kind === 220)) {
            return " '" + importPath(node.text) + "'";
        }
        if (node.kind === 120) {
            wasDeclared = true;
            return sourceFile.text.slice(node.getFullStart(), node.getStart());
        }
        if (node.kind === 230) {
            var path_3 = importPath(node.expression.text);
            return " require('" + path_3 + "')";
        }
    }
    function read(start, end) {
        var text = sourceFile.text.slice(start, end);
        if (end == null) {
            return text.replace(/\s+$/, '');
        }
        if (wasDeclared) {
            wasDeclared = false;
            return text.replace(/^\s+/, '');
        }
        return text;
    }
    var moduleText = processTree(sourceFile, replacer, read);
    var isEntry = options.entry === path;
    if (isEntry && options.isTypings) {
        return declareText(name, moduleText);
    }
    var moduleName = name + "/" + path_2.normalizeSlashes(path_2.relativeTo(tree.src, path_2.fromDefinition(path)));
    var declared = declareText(moduleName, moduleText);
    if (!isEntry) {
        return declared;
    }
    var importText = isES6Export ?
        "export * from '" + moduleName + "';" :
        "import main = require('" + moduleName + "');" + os_1.EOL + "export = main;";
    return "" + declared + os_1.EOL + declareText(name, importText);
}
function declareText(name, text) {
    return "declare module '" + name + "' {" + os_1.EOL + text + os_1.EOL + "}";
}
function processTree(sourceFile, replacer, read) {
    var code = '';
    var position = 0;
    function skip(node) {
        position = node.end;
    }
    function readThrough(node) {
        code += read(position, node.pos);
        position = node.pos;
    }
    function visit(node) {
        readThrough(node);
        var replacement = replacer(node);
        if (replacement != null) {
            code += replacement;
            skip(node);
        }
        else {
            ts.forEachChild(node, visit);
        }
    }
    visit(sourceFile);
    code += read(position);
    return code;
}
//# sourceMappingURL=compile.js.map